""" powermon / ports / usbport.py """
import asyncio
import logging
import os
import time
from glob import glob

from powermon.commands.command import Command
from powermon.commands.result import Result, ResultType
from powermon.dto.portDTO import PortDTO
from powermon.errors import ConfigError
from powermon.ports.abstractport import AbstractPort
from powermon.ports.porttype import PortType
from powermon.protocols import get_protocol_definition

log = logging.getLogger("USBPort")


class USBPort(AbstractPort):
    """ usb port object """
    @classmethod
    def from_config(cls, config=None):
        log.debug("building usb port. config:%s", config)
        path = config.get("path", "/dev/hidraw0")
        identifier = config.get("identifier")
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(path=path, protocol=protocol, identifier=identifier)

    def __init__(self, path, protocol, identifier) -> None:
        super().__init__(protocol=protocol)
        self.port_type = PortType.USB
        self.path = None
        self.port = None
        self.is_protocol_supported()
        # using glob to determine path(s)
        paths = glob(path)
        path_count = len(paths)
        match path_count:
            case 0:
                log.error("no matching paths found on this system for: %s", path)
                raise ConfigError(f"no matching paths found on this system for {path}")
            case 1:
                # only one valid result on this system
                self.path = paths[0]
            case _:
                # more than one valid path - so we need to determine which to use
                if identifier is None:
                    raise ConfigError("To use wildcard paths an identifier must be specified in the config file for the port")
                # need to build a command
                command = self.protocol.get_id_command()
                for _path in paths:
                    log.debug("Multiple paths - checking path: %s to see if it matches %s", _path, identifier)
                    self.path = _path
                    asyncio.run(self.connect())
                    res = asyncio.run(self.send_and_receive(command=command))
                    if not res.is_valid:
                        log.info("path: %s does not match for identifier: %s", _path, identifier)
                        asyncio.run(self.disconnect())
                        continue
                    if res.readings[0].data_value == str(identifier):
                        log.info("SUCCESS: path: %s matches for identifier: %s", _path, identifier)
                        return
                raise ConfigError(f"Multiple paths - none of {paths} match {identifier}")
        # end of multi-path logic

    def to_dto(self):
        dto = PortDTO(type="usb", path=self.path, protocol=self.protocol.to_dto())
        return dto

    def is_connected(self) -> bool:
        return self.port is not None

    async def connect(self) -> bool:
        if self.is_connected():
            log.debug("USBPort already connected")
            return True
        log.debug("USBPort connecting. path:%s, protocol:%s", self.path, self.protocol)
        try:
            self.port = os.open(self.path, os.O_RDWR | os.O_NONBLOCK)
            log.debug("USBPort port number $%s", self.port)
        except Exception as e:
            log.warning("Error openning usb port: %s", e)
            self.port = None
            self.error_message = e
        return self.is_connected()

    async def disconnect(self) -> None:
        log.debug("USBPort disconnecting: %i", self.port)
        if self.port is not None:
            os.close(self.port)
        self.port = None

    async def send_and_receive(self, command: Command) -> Result:
        if not self.is_connected():
            log.warning("USBPort not connected")
            return command.build_result(result_type=ResultType.ERROR, raw_response=b"USBPort not connected", protocol=self.protocol)
        response_line = bytes()

        # Send the command to the open usb connection
        full_command = command.full_command
        cmd_len = len(full_command)
        log.debug("length of to_send: %i", cmd_len)
        # for command of len < 8 it ok just to send
        # otherwise need to pack to a multiple of 8 bytes and send 8 at a time
        try:
            if cmd_len <= 8:
                # Send all at once
                log.debug("sending full_command in on shot")
                time.sleep(0.05)
                os.write(self.port, full_command)
            else:
                log.debug("multiple chunk send")
                chunks = [full_command[i:i + 8] for i in range(0, cmd_len, 8)]
                for chunk in chunks:
                    # pad chunk to 8 bytes
                    if len(chunk) < 8:
                        padding = 8 - len(chunk)
                        chunk += b'\x00' * padding
                    log.debug("sending chunk: %s", chunk)
                    time.sleep(0.05)
                    os.write(self.port, chunk)
            time.sleep(0.25)
            # Read from the usb connection
            # try to a max of 100 times
            for _ in range(100):
                # attempt to deal with resource busy and other failures to read
                time.sleep(0.15)
                r = os.read(self.port, 256)
                response_line += r
                    
                # Finished is \r is in byte_response
                if bytes([13]) in response_line:
                    # remove anything after the \r
                    response_line = response_line[: response_line.find(bytes([13])) + 1]
                    break
        except BrokenPipeError as e:
            log.debug("USB read error: %s", e)
        log.debug("usb response was: %s", response_line)
        # response = self.get_protocol().check_response_and_trim(response_line)
        result = command.build_result(raw_response=response_line, protocol=self.protocol)

        return result



""" powermon / ports / porttype.py """
from enum import auto
from strenum import LowercaseStrEnum


class PortType(LowercaseStrEnum):
    """ enumeration of supported / known port types """
    UNKNOWN = auto()
    TEST = auto()
    SERIAL = auto()
    USB = auto()
    BLE = auto()

    JKBLE = auto()
    MQTT = auto()
    VSERIAL = auto()
    DALYSERIAL = auto()
    ESP32 = auto()


""" powermon / protocols / jkserial.py """
import logging

import construct as cs

from powermon.commands.command_definition import CommandDefinition
from powermon.commands.reading_definition import ReadingType, ResponseType
from powermon.commands.result import ResultType
from powermon.commands.command import CommandType
from powermon.errors import InvalidResponse, CommandDefinitionMissing
from powermon.ports.porttype import PortType
from powermon.protocols.abstractprotocol import AbstractProtocol
from powermon.protocols.helpers import crc_jk232 as crc

log = logging.getLogger("jkserial")


# construct 'structures' to cover decoding of response packet
cell_details = cs.Struct("no" / cs.Byte, "voltage_mV" / cs.Int16ub)
all_data_response = cs.Struct(
    "stx" / cs.Const(b"NW"),
    "packet_length" / cs.Int16ub,
    "terminal_no" / cs.Bytes(4),
    "command_word" / cs.Byte,
    "frame_source" / cs.Enum(cs.Byte, BMS=0, Bluetooth=1, GPS=2, PC=3),
    "transport_type" / cs.Enum(cs.Byte, Response_Frame=1, BMS_Active_Upload=2),
    "_id" / cs.Const(b"\x79"),
    "data_length" / cs.Byte,
    "cell_count" / cs.Computed(cs.this.data_length // 3),
    "cell_array" / cs.Array(cs.this.cell_count, cell_details),
    "_id" / cs.Const(b"\x80"),
    "power_tube_temperature" / cs.Int16ub,
    "_id" / cs.Const(b"\x81"),
    "battery_box_temperature" / cs.Int16ub,
    "_id" / cs.Const(b"\x82"),
    "battery_temperature" / cs.Int16ub,
    "_id" / cs.Const(b"\x83"),
    "battery_voltage_10mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x84"),
    "battery_current" / cs.BitStruct("direction" / cs.Enum(cs.Bit, discharge=0, charge=1), "current" / cs.BitsInteger(15)),  # possible 10000 - cs.Int16ub (if \xc0 = 00)
    "_id" / cs.Const(b"\x85"),
    "battery_soc" / cs.Byte,
    "_id" / cs.Const(b"\x86"),
    "number_of_temp_sensors" / cs.Byte,
    "_id" / cs.Const(b"\x87"),
    "cycle_count" / cs.Int16ub,
    "_id" / cs.Const(b"\x89"),
    "total_cycle_capacity" / cs.Int32ub,
    "_id" / cs.Const(b"\x8a"),
    "battery_strings" / cs.Int16ub,
    "_id" / cs.Const(b"\x8b"),
    "warning_messages" / cs.BitStruct(
        'reserved_2' / cs.Flag,
        'reserved_1' / cs.Flag,
        '309_b_protection' / cs.Flag,
        '309_a_protection' / cs.Flag,
        'cell_under_voltage_alarm' / cs.Flag,
        'cell_over_voltage_alarm' / cs.Flag,
        'battery_low_temperature_alarm' / cs.Flag,
        'battery_box_over_temperature_alarm' / cs.Flag,
        'cell_pressure_difference_alarm' / cs.Flag,
        'discharge_over_current_alarm' / cs.Flag,
        'charging_over_current_alarm' / cs.Flag,
        'battery_over_temperature_alarm' / cs.Flag,
        'discharge_under_voltage_alarm' / cs.Flag,
        'charging_over_voltage_alarm' / cs.Flag,
        'mos_tube_over_temperature_alarm' / cs.Flag,
        'low_capacity_alarm' / cs.Flag),
    "_id" / cs.Const(b"\x8c"),
    "status_info" / cs.BitStruct(
        cs.Padding(12),
        'battery_down' / cs.Flag,
        'equalization_switching_state' / cs.Flag,
        'discharge_mos_on' / cs.Flag,
        'charge_mos_on' / cs.Flag,),
    # "test_optional" / cs.Optional(cs.Const(b"\x8d")),
    "_id" / cs.Const(b"\x8e"),
    "battery_overvoltage_protection_10mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x8f"),
    "battery_undervoltage_protection_10mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x90"),
    "cell_overvoltage_protection_mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x91"),
    "cell_overvoltage_protection_recovery_mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x92"),
    "cell_overvoltage_protection_delay_secs" / cs.Int16ub,
    "_id" / cs.Const(b"\x93"),
    "cell_undervoltage_protection_mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x94"),
    "cell_undervoltage_protection_recovery_mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x95"),
    "cell_undervoltage_protection_delay_secs" / cs.Int16ub,
    "_id" / cs.Const(b"\x96"),
    "cell_differential_protection_mV" / cs.Int16ub,
    "_id" / cs.Const(b"\x97"),
    "discharge_protection_current_A" / cs.Int16ub,
    "_id" / cs.Const(b"\x98"),
    "discharge_protection_current_delay_secs" / cs.Int16ub,
    "_id" / cs.Const(b"\x99"),
    "charge_protection_current_A" / cs.Int16ub,
    "_id" / cs.Const(b"\x9a"),
    "charge_protection_current_delay_secs" / cs.Int16ub,
    "_id" / cs.Const(b"\x9b"),
    "equalization_starting_voltage_mV" / cs.Int16ub,

    "rest" / cs.Bytes(93),  # TODO: add rest of packet

    "_id" / cs.Const(b"\xb7"),
    "software_id" / cs.Bytes(15),
    "_id" / cs.Const(b"\xb8"),
    "start_calibration" / cs.Enum(cs.Byte, StartCalibration=1, CloseCalibration=0),
    "_id" / cs.Const(b"\xb9"),
    "battery_capacity_Ah" / cs.Int32ub,
    "_id" / cs.Const(b"\xba"),
    "manufacturer_name" / cs.Bytes(24),
    "_id" / cs.Const(b"\xc0"),
    "protocol_version" / cs.Int8ub,
    "record_number" / cs.Bytes(4),
    "end_of_identity" / cs.Const(b"h"),
    "checksum" / cs.Bytes(4)
)

COMMANDS = {
    "all_data": {
        "name": "all_data",
        "description": "Get All BMS Data",
        "help": " -- Get All BMS Data",
        "construct": all_data_response,
        "command_type": CommandType.SERIAL_READONLY,
        "command_code": "00",
        "result_type": ResultType.CONSTRUCT,
        "reading_definitions": [
            {"index": "cell_count", "description": "Cell Count"},
            {"index": "cell_1_voltage", "description": "Cell 1 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_2_voltage", "description": "Cell 2 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_3_voltage", "description": "Cell 3 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_4_voltage", "description": "Cell 4 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_5_voltage", "description": "Cell 5 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_6_voltage", "description": "Cell 6 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_7_voltage", "description": "Cell 7 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_8_voltage", "description": "Cell 8 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_9_voltage", "description": "Cell 9 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_10_voltage", "description": "Cell 10 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_11_voltage", "description": "Cell 11 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_12_voltage", "description": "Cell 12 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_13_voltage", "description": "Cell 13 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_14_voltage", "description": "Cell 14 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_15_voltage", "description": "Cell 15 Voltage", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_16_voltage", "description": "Cell 16 Voltage", "reading_type": ReadingType.MILLI_VOLTS},

            {"index": "power_tube_temperature", "description": "MOS Temperature", "reading_type": ReadingType.TEMPERATURE},
            {"index": "battery_box_temperature", "description": "Battery Box Temperature", "reading_type": ReadingType.TEMPERATURE},
            {"index": "battery_temperature", "description": "Battery Temperature", "reading_type": ReadingType.TEMPERATURE},
            {"index": "number_of_temp_sensors", "description": "Number of Temperature Sensors"},

            {"index": "battery_voltage_10mV", "description": "Battery Voltage", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "direction", "description": "direction", "response_type": ResponseType.STRING},
            {"index": "current", "description": "Battery Current", "reading_type": ReadingType.CURRENT, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "protocol_version", "description": "Protocol Version", "response_type": ResponseType.STRING},
            {"index": "battery_soc", "description": "Battery State of Charge", "reading_type": ReadingType.PERCENTAGE},
            {"index": "cycle_count", "description": "Battery Cycle Count"},
            {"index": "battery_strings", "description": "Battery Strings"},
            {"index": "total_cycle_capacity", "description": "Total Cycle Count Capacity", "reading_type": ReadingType.ENERGY},
            {"index": "battery_capacity_Ah", "description": "Battery Capacity", "reading_type": ReadingType.ENERGY},

            {"index": "software_id", "description": "Software ID", "response_type": ResponseType.BYTES},
            {"index": "manufacturer_name", "description": "Manufacturer Name", "response_type": ResponseType.TEMPLATE_BYTES, "format_template" : "r.removeprefix('Input Userda')"},

            {"index": "charge_mos_on", "description": "Charge MOS On", "response_type": ResponseType.BOOL},
            {"index": "discharge_mos_on", "description": "Discharge MOS On", "response_type": ResponseType.BOOL},
            {"index": "equalization_switching_state", "description": "Equalization Switching State", "response_type": ResponseType.BOOL},
            {"index": "equalization_starting_voltage_mV", "description": "Equalization Starting Voltage Setting", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "battery_down", "description": "Battery Down", "response_type": ResponseType.BOOL},
            {"index": "start_calibration", "description": "Start Calibration", "response_type": ResponseType.STRING},
            {"index": "warning_messages", "description": "Warning Messages", "reading_type": ReadingType.HEX_STR, "response_type": ResponseType.INT},

            {"index": "low_capacity_alarm", "description": "Low Capacity Alarm", "response_type": ResponseType.BOOL},
            {"index": "mos_tube_over_temperature_alarm", "description": "MOS Tube Over Temperature Alarm", "response_type": ResponseType.BOOL},
            {"index": "charging_over_voltage_alarm", "description": "charging_over_voltage_alarm", "response_type": ResponseType.BOOL},
            {"index": "discharge_under_voltage_alarm", "description": "discharge_under_voltage_alarm", "response_type": ResponseType.BOOL},
            {"index": "battery_over_temperature_alarm", "description": "battery_over_temperature_alarm", "response_type": ResponseType.BOOL},
            {"index": "charging_over_current_alarm", "description": "charging_over_current_alarm", "response_type": ResponseType.BOOL},
            {"index": "discharge_over_current_alarm", "description": "discharge_over_current_alarm", "response_type": ResponseType.BOOL},
            {"index": "cell_pressure_difference_alarm", "description": "cell_pressure_difference_alarm", "response_type": ResponseType.BOOL},
            {"index": "battery_box_over_temperature_alarm", "description": "battery_box_over_temperature_alarm", "response_type": ResponseType.BOOL},
            {"index": "battery_low_temperature_alarm", "description": "battery_low_temperature_alarm", "response_type": ResponseType.BOOL},
            {"index": "cell_over_voltage_alarm", "description": "cell_over_voltage_alarm", "response_type": ResponseType.BOOL},
            {"index": "cell_under_voltage_alarm", "description": "cell_under_voltage_alarm", "response_type": ResponseType.BOOL},
            {"index": "309_a_protection", "description": "309_a_protection", "response_type": ResponseType.BOOL},
            {"index": "309_b_protection", "description": "309_b_protection", "response_type": ResponseType.BOOL},

            {"index": "battery_overvoltage_protection_10mV", "description": "Battery Overvoltage Protection Setting", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "battery_undervoltage_protection_10mV", "description": "Battery Undervoltage Protection Setting", "reading_type": ReadingType.VOLTS, "response_type": ResponseType.TEMPLATE_INT, "format_template": "r/100"},
            {"index": "cell_overvoltage_protection_mV", "description": "Cell Overvoltage Protection Setting", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_overvoltage_protection_recovery_mV", "description": "Cell Overvoltage Protection Recovery Setting", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_overvoltage_protection_delay_secs", "description": "Cell Overvoltage Protection Recovery Delay Setting", "reading_type": ReadingType.TIME_SECONDS},
            {"index": "cell_undervoltage_protection_mV", "description": "Cell Undervoltage Protection Setting", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_undervoltage_protection_recovery_mV", "description": "Cell Undervoltage Protection Recovery Setting", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "cell_undervoltage_protection_delay_secs", "description": "Cell Undervoltage Protection Recovery Delay Setting", "reading_type": ReadingType.TIME_SECONDS},
            {"index": "cell_differential_protection_mV", "description": "Cell Voltage Difference Protection Setting", "reading_type": ReadingType.MILLI_VOLTS},
            {"index": "charge_protection_current_A", "description": "Charge Current Protection Setting", "reading_type": ReadingType.CURRENT},
            {"index": "charge_protection_current_delay_secs", "description": "Charge Current Protection Delay Setting", "reading_type": ReadingType.TIME_SECONDS},
            {"index": "discharge_protection_current_A", "description": "Discharge Current Protection Setting", "reading_type": ReadingType.CURRENT},
            {"index": "discharge_protection_current_delay_secs", "description": "Discharge Current Protection Delay Setting", "reading_type": ReadingType.TIME_SECONDS},

            # fields to ignore
            {"index": "stx", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "_id", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "packet_length", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "terminal_no", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "command_word", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "frame_source", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "transport_type", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "data_length", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "record_number", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "end_of_identity", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "checksum", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "reserved_1", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "reserved_2", "description": "", "reading_type": ReadingType.IGNORE},
            {"index": "rest", "description": "", "reading_type": ReadingType.IGNORE},

            # ["BigHex2Short:(r&0x7FFF)/100*(((r&0x8000)>>15)*2-1)", 2, "Battery_Current", "A"],

        ],
        "test_responses": [
            b'NW\x01\x1b\x00\x00\x00\x00\x03\x00\x01y*\x01\x0f\x90\x02\x0f\x91\x03\x0f\x94\x04\x0f\x8e\x05\x0f\x92\x06\x0f\x91\x07\x0f\x91\x08\x0f\x91\t\x0f\x93\n\x0f\x8e\x0b\x0f\x91\x0c\x0f\x90\r\x0f\x90\x0e\x0f\x8d\x80\x00!\x81\x00\x1c\x82\x00\x1e\x83\x15\xca\x84\x81\xc5\x85d\x86\x02\x87\x00\x19\x89\x00\x00\x16\xda\x8a\x00\x0e\x8b\x00\x00\x8c\x00\x03\x8e\x16\xb2\x8f\x10\xf4\x90\x106\x91\x10\x04\x92\x00\x05\x93\x0c\x1c\x94\x0c\x80\x95\x00\x05\x96\x01,\x97\x00n\x98\x01,\x99\x00U\x9a\x00\x1e\x9b\x0b\xb8\x9c\x002\x9d\x01\x9e\x00Z\x9f\x00F\xa0\x00d\xa1\x00d\xa2\x00\x14\xa3\x00<\xa4\x00<\xa5\x00\x01\xa6\x00\x03\xa7\xff\xec\xa8\xff\xf6\xa9\x0e\xaa\x00\x00\x00\xea\xab\x01\xac\x01\xad\x047\xae\x01\xaf\x01\xb0\x00\n\xb1\x14\xb2123456\x00\x00\x00\x00\xb3\x00\xb4Input Us\xb52306\xb6\x00\x01\x82\xe3\xb711.XW_S11.261__\xb8\x00\xb9\x00\x00\x00\xea\xbaInput UserdaJK_B1A20S15P\xc0\x01\x00\x00\x00\x00h\x00\x00Q\xd6',
            bytes.fromhex("4e 57 00 fd 00 00 00 00 06 00 01 79 0c 01 0d 06 02 0d 06 03 0d 07 04 0d 07 80 00 10 81 00 0e 82 00 0d 83 05 35 84 00 00 85 62 86 02 87 00 00 89 00 00 00 05 8a 00 04 8b 00 03 8c 00 03 8e 05 a0 8f 04 10 90 0e 10 91 0d de 92 00 05 93 0a 28 94 0a 5a 95 00 05 96 01 2c 97 00 78 98 00 1e 99 00 3c 9a 00 1e 9b 0d 48 9c 00 05 9d 01 9e 00 50 9f 00 41 a0 00 64 a1 00 64 a2 00 14 a3 00 32 a4 00 37 a5 00 03 a6 00 08 a7 ff ec a8 ff f6 a9 04 aa 00 00 01 31 ab 01 ac 01 ad 03 7e ae 01 af 00 b0 00 0a b1 14 b2 35 33 31 34 00 00 00 00 00 00 b3 00 b4 49 6e 70 75 74 20 55 73 b5 32 33 31 32 b6 00 00 36 a6 b7 31 31 2e 58 57 5f 53 31 31 2e 32 31 48 5f 5f b8 00 b9 00 00 01 31 ba 49 6e 70 75 74 20 55 73 65 72 64 61 45 64 64 69 65 42 6c 75 65 42 4d 53 c0 01 00 00 00 00 68 00 00 44 6f"),
        ],
    },
    "battery_voltage": {
        "name": "battery_voltage",
        "description": "Get the battery voltage",
        "help": " -- Get the battery voltage",
        # "construct": balancer_data_response,
        "command_type": CommandType.JKSERIAL_READ,
        "command_code": "83",
        "result_type": ResultType.SINGLE,
        "reading_definitions": [{"description": "Battery Voltage"}],
        "test_responses": []}
}


class JkSerial(AbstractProtocol):
    """ JKBMS TTL serial communication protocol handler """
    def __str__(self):
        return "JKBMS TTL serial communication protocol handler"

    def __init__(self) -> None:
        super().__init__()
        self._protocol_id = b"JKSERIAL"
        self.add_command_definitions(COMMANDS)
        self.add_supported_ports([PortType.SERIAL])
        self.check_definitions_count(expected=2)

    def check_valid(self, response: str, command_definition: CommandDefinition = None) -> bool:
        """ check response is valid """
        log.debug("check valid for %s, definition: %s", response, command_definition)
        if response is None:
            raise InvalidResponse("Response is None")
        if len(response) <= 3:
            raise InvalidResponse("Response is too short")
        # check start bytes = 0x4e57 / NW
        if response[0:2] != b"NW":
            raise InvalidResponse(f"Start bytes ({response[0:2]} not 'NW')")
        return True

    def check_crc(self, response: str, command_definition: CommandDefinition = None):
        return True

    def trim_response(self, response: str, command_definition: CommandDefinition = None) -> str:
        """ Remove extra characters from response """
        log.debug("trim %s, definition: %s", response, command_definition)
        return response

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different
        """
        log.info("Using protocol: %s with %i commands", self.protocol_id, len(self.command_definitions))
        command_defn = self.get_command_definition(command)
        # raise error is no command_defn found
        if command_defn is None:
            raise CommandDefinitionMissing(f"No definition found in JKSERIAL for {command}")

        # command byte: 0x01 (activation), 0x02 (write), 0x03 (read), 0x05 (password), 0x06 (read all)
        match command_defn.command_type:
            case CommandType.SERIAL_READONLY:
                command_byte = 0x06
            case CommandType.JKSERIAL_ACTIVATION:
                command_byte = 0x01
            case CommandType.JKSERIAL_SETTER:
                command_byte = 0x02
            case _:
                command_byte = 0x03

        # Read basic information and status
        # full command is 21 bytes long
        cmd = bytearray(21)
        command_code = int(command_defn.command_code, 16)

        # start bit  0x4E
        cmd[0] = 0x4E                         # start sequence
        cmd[1] = 0x57                         # start sequence
        cmd[2] = 0x00                         # data length lb
        cmd[3] = 0x13                         # data length hb
        cmd[4] = 0x00                         # bms terminal number
        cmd[5] = 0x00                         # bms terminal number
        cmd[6] = 0x00                         # bms terminal number
        cmd[7] = 0x00                         # bms terminal number
        cmd[8] = command_byte
        cmd[9] = 0x03                         # frame source: 0x00 (bms), 0x01 (bluetooth), 0x02 (gps), 0x03 (computer)
        cmd[10] = 0x00                        # frame type: 0x00 (read data), 0x01 (reply frame), 0x02 (BMS active upload)
        cmd[11] = command_code                # register: 0x00 (read all registers), 0x8E...0xBF (holding registers)
        cmd[12] = 0x00                        # record number
        cmd[13] = 0x00                        # record number
        cmd[14] = 0x00                        # record number
        cmd[15] = 0x00                        # record number
        cmd[16] = 0x68                        # end sequence
        cmd[17] = 0x00                        # crc unused
        cmd[18] = 0x00                        # crc unused
        crc_high, crc_low = crc(cmd[0:17])
        cmd[19] = crc_high
        cmd[20] = crc_low

        log.debug("cmd with crc: %s", cmd)
        return cmd

    def split_response(self, response: str, command_definition: CommandDefinition = None) -> list:
        """ split response into individual items, return as ordered list or list of tuples """
        result_type = getattr(command_definition, "result_type", None)
        log.debug("jkserial splitting %s, result_type %s", response, result_type)
        # build a list of (index, value) tuples, after parsing with a construct
        responses = []
        # parse with construct
        result = command_definition.construct.parse(response)
        # print(result)
        for x in result:
            # if x == "cell_array":
            #     print(x, result[x])
            match type(result[x]):
                case cs.ListContainer:
                    # print(result[x])
                    if x == "cell_array":
                        for _container in result[x]:
                            key = f"cell_{_container['no']}_voltage"
                            value = _container['voltage_mV']
                            responses.append((key, value))
                case cs.Container:
                    for y in result[x]:
                        if y != "_io":
                            key = y
                            value = result[x][y]
                            responses.append((key, value))
                case _:
                    if x != "_io":
                        key = x
                        value = result[x]
                        responses.append((key, value))
            # case _:
            #     responses = response.split()
        log.debug("responses: '%s'", responses)
        return responses



""" powermon / ports / serialport.py """
import asyncio
import logging
import time
from glob import glob

import serial

from powermon.commands.command import Command, CommandType
from powermon.commands.result import Result
from powermon.dto.portDTO import PortDTO
from powermon.errors import ConfigError
from powermon.ports.abstractport import AbstractPort
from powermon.ports.porttype import PortType
from powermon.protocols import get_protocol_definition

log = logging.getLogger("SerialPort")


class SerialPort(AbstractPort):
    """ serial port object - normally a usb to serial adapter """

    def __str__(self):
        return f"SerialPort: {self.path=}, {self.baud=}, protocol:{self.protocol}, {self.serial_port=}, {self.error_message=}"

    @classmethod
    def from_config(cls, config=None):
        log.debug("building serial port. config:%s", config)
        path = config.get("path", "/dev/ttyUSB0")
        baud = config.get("baud", 2400)
        identifier = config.get("identifier")
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(path=path, baud=baud, protocol=protocol, identifier=identifier)

    def __init__(self, path, baud, protocol, identifier) -> None:
        super().__init__(protocol=protocol)
        self.port_type = PortType.SERIAL
        self.is_protocol_supported()
        self.path = None
        self.baud = baud
        self.serial_port = None
        # self.identifier = identifier
        # using glob to determine path(s)
        paths = glob(path)
        path_count = len(paths)
        match path_count:
            case 0:
                log.error("no matching paths found on this system for: %s", path)
                raise ConfigError(f"no matching paths found on this system for {path}")
            case 1:
                # only one valid result on this system
                self.path = paths[0]
            case _:
                # more than one valid path - so we need to determine which to use
                if identifier is None:
                    raise ConfigError("To use wildcard paths an identifier must be specified in the config file for the port")
                # need to build a command
                command = self.protocol.get_id_command()
                for _path in paths:
                    log.debug("Multiple paths - checking path: %s to see if it matches %s", _path, identifier)
                    self.path = _path
                    asyncio.run(self.connect())
                    res = asyncio.run(self.send_and_receive(command=command))
                    if not res.is_valid:
                        log.debug("path: %s does not match for identifier: %s", _path, identifier)
                        continue
                    # print(res.readings[0])
                    # print(res.readings[0].data_value)
                    # print(res.readings[0].data_value == identifier)
                    if res.readings[0].data_value == identifier:
                        log.info("SUCCESS: path: %s matches for identifier: %s", _path, identifier)
                        return
                raise ConfigError(f"Multiple paths - none of {paths} match {identifier}")
        # end of multi-path logic

    def to_dto(self) -> PortDTO:
        dto = PortDTO(type="serial", path=self.path, baud=self.baud, protocol=self.protocol.to_dto())
        return dto

    def is_connected(self):
        return self.serial_port is not None and self.serial_port.is_open

    async def connect(self) -> int:
        log.debug("usbserial port connecting. path:%s, baud:%s", self.path, self.baud)
        try:
            self.serial_port = serial.Serial(port=self.path, baudrate=self.baud, timeout=1, write_timeout=1)
            log.debug(self.serial_port)
        except ValueError as e:
            log.error("Incorrect configuration for serial port: %s", e)
            self.error_message = str(e)
            self.serial_port = None
        except serial.SerialException as e:
            log.error("Error opening serial port: %s", e)
            self.error_message = str(e)
            self.serial_port = None
        return self.is_connected()

    async def disconnect(self) -> None:
        log.debug("usbserial port disconnecting")
        if self.serial_port is not None:
            self.serial_port.close()
        self.serial_port = None

    async def send_and_receive(self, command: Command) -> Result:
        full_command = command.full_command
        response_line = None
        log.info("port: %s, full_command: %s", self.serial_port, full_command)
        if not self.is_connected():
            raise RuntimeError("Serial port not open")
        try:
            log.debug("Executing command via usbserial...")
            self.serial_port.reset_input_buffer()
            self.serial_port.reset_output_buffer()
            # Process i/o differently depending on command type
            command_defn = command.command_definition
            match command_defn.command_type:
                case CommandType.VICTRON_LISTEN:
                    # this command type doesnt need to send a command, it just listens on the serial port
                    _lines = 30
                    log.debug("case: CommandType.VICTRON_LISTEN, listening for %i lines", _lines)
                    response_line = b""
                    for _ in range(_lines):
                        _response = self.serial_port.read_until(b"\n")
                        response_line += _response
                case CommandType.SERIAL_READONLY:
                    # read until no more data
                    log.debug("CommandType.SERIAL_READONLY")
                    response_line = b""
                    while True:
                        await asyncio.sleep(0.2)  # give serial port time to receive the data
                        to_read = self.serial_port.in_waiting
                        log.debug("bytes waiting: %s", to_read)
                        if to_read == 0:
                            break
                        # got some data to read
                        response_line += self.serial_port.read(to_read)
                case CommandType.SERIAL_READ_UNTIL_DONE:
                    # this case reads until no more to read or timeout
                    log.debug("case: CommandType.SERIAL_READ_UNTIL_DONE")
                    response_line = b""
                    self.serial_port.timeout = 0.5
                    self.serial_port.write_timeout = 1
                    self.serial_port.reset_input_buffer()
                    self.serial_port.reset_output_buffer()
                    c = self.serial_port.write(full_command)
                    log.debug("wrote %s bytes", c)
                    self.serial_port.flush()
                    # read until no more data
                    while True:
                        # await asyncio.sleep(0.5)  # give serial port time to receive the data
                        time.sleep(0.5)
                        to_read = self.serial_port.in_waiting
                        log.debug("bytes waiting %s", to_read)
                        if to_read == 0:
                            break
                        # got some data to read
                        response_line += self.serial_port.read(to_read)
                case _:
                    # default processing
                    self.serial_port.reset_input_buffer()
                    self.serial_port.reset_output_buffer()
                    c = self.serial_port.write(full_command)
                    log.debug("Default serial s&r. Wrote %i bytes", c)
                    self.serial_port.flush()
                    time.sleep(0.3)  # give serial port time to receive the data
                    response_line = self.serial_port.read_until(b"\r")
            log.info("serial response was: %s", response_line)
            # response = self.get_protocol().check_response_and_trim(response_line)
            result = command.build_result(raw_response=response_line, protocol=self.protocol)
            return result
        except Exception as e:
            log.warning("Serial read error: %s", e)
            result.error = True
            result.error_messages.append(f"Serial read error {e}")
            self.disconnect()
            return result



""" powermon / protocols / helpers.py """
import logging

log = logging.getLogger("protocol_helpers")


def crc_jk232(byte_data):
    """
    Generate JK RS232 / RS485 CRC
    - 2 bytes, the verification field is "command code + length byte + data segment content",
    the verification method is thesum of the above fields and then the inverse plus 1, the high bit is in the front and the low bit is in the back.
    """
    crc = 0
    for b in byte_data:
        crc += b
    crc_low = crc & 0xFF
    crc_high = (crc >> 8) & 0xFF
    return [crc_high, crc_low] 

def crc_pi30(data_bytes):
    """
    Calculates CRC for supplied data_bytes
    """
    # assert type(byte_cmd) == bytes
    log.debug("Calculating CRC for %s", data_bytes)

    crc = 0
    da = 0
    crc_ta = [
        0x0000,
        0x1021,
        0x2042,
        0x3063,
        0x4084,
        0x50A5,
        0x60C6,
        0x70E7,
        0x8108,
        0x9129,
        0xA14A,
        0xB16B,
        0xC18C,
        0xD1AD,
        0xE1CE,
        0xF1EF,
    ]

    for c in data_bytes:
        # log.debug('Encoding %s', c)
        if isinstance(c, str):
            c = ord(c)
        da = ((crc >> 8) & 0xFF) >> 4
        crc = (crc << 4) & 0xFFFF

        index = da ^ (c >> 4)
        crc ^= crc_ta[index]

        da = ((crc >> 8) & 0xFF) >> 4
        crc = (crc << 4) & 0xFFFF

        index = da ^ (c & 0x0F)
        crc ^= crc_ta[index]

    crc_low = crc & 0xFF
    crc_high = (crc >> 8) & 0xFF

    if crc_low == 0x28 or crc_low == 0x0D or crc_low == 0x0A or crc_low == 0x00:
        crc_low += 1
    if crc_high == 0x28 or crc_high == 0x0D or crc_high == 0x0A or crc_high == 0x00:
        crc_high += 1

    crc = crc_high << 8
    crc += crc_low

    log.debug(f"Generated CRC {crc_high:#04x} {crc_low:#04x} {crc:#06x}")
    return [crc_high, crc_low]

def victron_checksum(byte_data):
    """
    Generate VE Direct HEX Checksum
    - sum of byteData + checksum = 0x55
    """
    checksum = 0x55
    for b in byte_data:
        checksum -= b
    checksum = checksum & 0xFF
    return checksum

    
